# Copyright (c) 2008,2009 SFB/TR-55
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

""" Convenience python module.
    It is meant to make the use of the frontend daemon easier.
"""

import array
import string
import getopt
import sys
import os
import socket

DAEMON_DEFAULT_PORT     = 17000
DAEMON_DEFAULT_CONTROL  = 17001
RootcardPattern = "rc-%02i-%02i"
NodecardPattern = "nc-%02i-%02i"
HemauPattern = "src-%i"

# Definition for SPI handling.
# See @a feenSPIShift for details.
FEEN_CS_AUTO_OFF = 0xCCC2

# Definition for SPI handling.
# See @a feenSPIShift for details.
FEEN_CS_AUTO_1 = 0xCCC1

# Definition for SPI handling.
# See @a feenSPIShift for details.
FEEN_CS_AUTO_0 = 0xCCC0

# Rootcard scans for a new packet.
# See @a feenStatus for details.
FEEN_SCANNING = 0x88000001 # /* keep in sync with ROBPROT_SCANNING */

# Rootcard processes internal command.
# See @a feenStatus for details.
FEEN_INTER_WIP = 0x88000002 # /* keep in sync with ROBPROT_INTER_WIP */

# Rootcard processes external command.
# See @a feenStatus for details.
FEEN_EXTERN_WIP = 0x88000003 # /* keep in sync with ROBPROT_EXTERN_WIP */

# Rootcard is idle.
# See @a feenStatus for details.
FEEN_IDLE = 0x88000004 # /* keep in sync with ROBPROT_IDLE */

FEEN_SECTOR_SIZE_FLASH_INTERNAL = 2048
FEEN_SECTOR_SIZE_FLASH_EXTERNAL = 65536
FEEN_PAGE_SIZE_FLASH = 256


def qlistind_(indent):
    """ Helper function to indent output.
    """
    ret = ''
    for i in range(0, indent):
        ret = ret + '  '
    return ret

class qdata:
    """ Wrapper class for binary data.
        If present, hex is a list of space-separated octets in 
        hex format, the leading 0x.. is optional
        e.g,:
        d = qdata('', hex='ff 17 e3')
    """
    def __init__(self, x, hex=''):
        self.data = x
        if len(hex) > 0:
            octets = string.split(hex)
            for o in octets:
                self.data += '%c' % (string.atoi(o, 16))
    def hexdump(self, indent):
        retstr = ''
        retstr += qlistind_(indent)
        arr = array.array('B', self.data)
        lm = 0
        for x in arr:
            lm = lm + 1
            retstr += ('%02x' % (x))
            if (lm % 16):
                retstr += ' '
            else:
                retstr += '\n'
                retstr += qlistind_(indent)
        return retstr

def py2qlist_(py, indent):
    """ Helper function
    """
    ret = ''
    ret += qlistind_(indent)
    if isinstance(py, bool):
        if py:
            return ret + '<boolean><true/></boolean>\n'
        else:
            return ret + '<boolean><false/></boolean>\n'        
    elif isinstance(py, int):
        return ret + '<integer>' + ('%i' % (py)) + '</integer>' + '\n'
    elif isinstance(py, float):
        return ret + '<real>' + ('%f' % (py)) + '</read>' + '\n'
    elif isinstance(py, str) or isinstance(py, unicode):
        return ret + '<string>' + py + '</string>' + '\n'
    elif isinstance(py, qdata):
        ret += '<data>\n'
        ret += py.hexdump(indent + 1)
        ret += '\n'
        ret += qlistind_(indent)
        ret += '</data>\n'
        return ret
    elif isinstance(py, list):
        ret += '<array>\n'
        for i in py:
            ret += py2qlist_(i, indent + 1)
        ret += qlistind_(indent)
        ret += '</array>\n'
        return ret
    elif isinstance(py, dict):
        ret += '<dict>\n'
        for (k,v) in py.items():
            ret += qlistind_(indent+1)
            ret += '<QPair>\n'
            ret += qlistind_(indent+2)
            ret += '<key>' + k + '</key>' + '\n'
            ret += qlistind_(indent+2)
            ret += '<value>\n'
            ret += py2qlist_(v, indent + 3)
            ret += qlistind_(indent+2)
            ret += '</value>\n'
            ret += qlistind_(indent+1)
            ret += '</QPair>\n'
        ret += qlistind_(indent)
        ret += '</dict>\n'
        return ret


def py2qlist(ql):
    """ Converts a python data structure containing
        bool, dict, list, integer, qdata, and float values
        to a QList XML document.
    """
    return '<QList>\n' + py2qlist_(ql,1) + '</QList>\n'

# check if DOM is suited better for parsing XML QList objects

import xml.dom.minidom

def qlist2py_(d):
    """ Helper function.
    """
    ret = ''
    if d.nodeName == 'array':
        ret = []
        for c in d.childNodes:
            if c.nodeType == c.ELEMENT_NODE:
                ret.append(qlist2py_(c))

    if d.nodeName == 'dict':
        ret = {}
        for c in d.childNodes:
            if c.nodeType == c.ELEMENT_NODE:
                ret.update(qlist2py_(c))

    if d.nodeName == 'QPair':
        key = ''
        value = ''
        for c in d.childNodes:
            if c.nodeType == c.ELEMENT_NODE:
                if c.nodeName == 'key':
                    key = c.childNodes[0].nodeValue
                if c.nodeName == 'value':
                    for v in c.childNodes:
                        if v.nodeType == v.ELEMENT_NODE:
                            value = qlist2py_(v)
                            break
        ret = [(key, value)]

    if d.nodeName == 'string':
        if not len(d.childNodes) > 0:
            ret = ''
        else:
            ret = d.childNodes[0].nodeValue

    if d.nodeName == 'boolean':
        for c in d.childNodes:
            if c.nodeName == 'true':
                ret = True
            if c.nodeName == 'false':
                ret = False

    if d.nodeName == 'integer':
        ret = string.atoi(d.childNodes[0].nodeValue)

    if d.nodeName == 'real':
        ret = string.atof(d.childNodes[0].nodeValue)

    if d.nodeName == 'data':
        ret = qdata('')
        if not len(d.childNodes) > 0:
            ret = qdata('')
        else:
            hx = d.childNodes[0].nodeValue.split()
            for h in hx:
                ix = string.atoi(h, 16)
                ret.data += '%c' % (ix)

    return ret


def qlist2py(xmldoc):
    """ Converts a QList XML document back to a python data
        structure containing list, dict, qdata, int, float,
        and bool types.
        Relies on xml.dom.minidom
    """
    doc = xml.dom.minidom.parseString(xmldoc)
    if doc.nodeType != doc.DOCUMENT_NODE:
        raise Exception('no document')
    roots = doc.childNodes
    if len(roots) != 1:
        raise Exception('qlist needs exactly one child node')
    if roots[0].nodeName != 'QList':
        raise Exception('this is not a qlist')
    subroots = roots[0].childNodes
    foundsomething = False
    for s in subroots:
        if s.nodeType == s.TEXT_NODE:
            continue
        if s.nodeType == s.ELEMENT_NODE:
            if s.nodeName in ('dict','array','string','integer','real','boolean'):
                foundsomething = True
                return qlist2py_(s)
    return ''

def list2array(list):
    """ Convert commonly used expressions for ranges
        of integers like:
            1,2,7,8-11,14-21
        to a list of int.
    """
    ret = []
    cs = string.split(list, ',')
    for v in cs:
        if len(v) == 0: continue
        r = string.split(v, '-')
        if len(r) == 1:
            ret.append(string.atoi(r[0],0))
        elif len(r) == 2:
            ret += range(string.atoi(r[0],0),\
                         string.atoi(r[1],0) + 1)
        else:
            raise Exception('not a proper range')

    return ret

def argv2rootcards(arguments=sys.argv[1:], optstring='b:r:n:'):
    """ Convenience function.
        Converts a list of strings (e.g., the command line arguments)
        to a list of rootcards.
        Example: if the command line is:  -b 2-3 -r 0
        the return value is:
        ['rc-02-00'], ['rc-03-00']
    """
    (opts, nonopts) = getopt.getopt(arguments, optstring)
    backlist = []
    rootlist = []
    rootips = []
    for (opt, arg) in opts:
        if opt == '-b':
            backlist = list2array(arg)
        if opt == '-r':
            rootlist = list2array(arg)

            for b in backlist:
                for r in rootlist:
                    rootips.append( RootcardPattern % (b, r))

    return rootips

def argv2nodecards(arguments=sys.argv[1:], optstring='b:r:n:'):
    """ Convenience function.
        Converts a list of strings (e.g., the command line arguments)
        to a list of nodecards.
        Example: if the command line is:  -b 2-3 -n 31,0
        the return value is:
        ['nc-02-31', 'nc-02-00', 'nc-03-31', 'nc-03-00']
    """
    (opts, nonopts) = getopt.getopt(arguments, optstring)
    backlist = []
    nodelist = []
    nodeips = []
    for (opt, arg) in opts:
        if opt == '-b':
            backlist = list2array(arg)
        if opt == '-n':
            nodelist = list2array(arg)

            for b in backlist:
                for n in nodelist:
                    nodeips.append( NodecardPattern % (b, n))

    return nodeips

def argv2rootnodedoublets(arguments=sys.argv[1:], optstring='b:r:n:'):
    """ Convenience function.
        Converts command line arguments to a list of tuples,
        where the first element of the touple is a string
        with the rootcard IP, and the second element of the touble is
        an int with the nodecard number as seen on the RS485 bus.
        Example: if the command line is:  -b 2-3 -n 31,0
        the return value is:
        [('rc-02-01', 16), ('rc-02-00', 0), ('rc-03-01', 15), ('rc-03-00', 0)]
    """
    (opts, nonopts) = getopt.getopt(arguments, optstring)
    backlist = []
    nodelist = []
    doublets = []
    for (opt, arg) in opts:
        if opt == '-b':
            backlist = list2array(arg)
        if opt == '-n':
            nodelist = list2array(arg)
            for b in backlist:
                for n in nodelist:
                    if n >= 16:
                        doublets.append((RootcardPattern % (b, 1), n-16))
                    else:
                        doublets.append((RootcardPattern % (b, 0), n))

    return doublets

def argv2rootnodetree(arguments=sys.argv[1:], optstring='b:r:n:'):
    """ Convenience function.
        Converts the command line to a dict, there the keys
        are the rootcard IPs the values are lists of int,
        containing the nodecard IDs as seen on the RS485 bus.
        Example: if the command line is: -b 2-3 -n 31,30,0-1
        the return value is:
        {'rc-02-01':[14,15], 'rc-02-00':[0,1], 'rc-03-01':[14,16], 'rc-03-00':[0,1]}
    """
    (opts, nonopts) = getopt.getopt(arguments, optstring)
    backlist = []
    nodelist = []
    branch = {}
    for (opt, arg) in opts:
        if opt == '-b':
            backlist = list2array(arg)
        if opt == '-n':
            nodelist = list2array(arg)
            for b in backlist:
                for n in nodelist:
                    if n >= 16:
                        key = RootcardPattern % (b, 1)
                        if not key in branch.keys():
                            branch.update([(key,[])])
                        leaves = branch[key]
                        leaves.append(n-16)
                    else:
                        key = RootcardPattern % (b, 0)
                        if not key in branch.keys():
                            branch.update([(key,[])])
                        leaves = branch[key]
                        leaves.append(n)
    return branch

def daemonize(host, port=DAEMON_DEFAULT_PORT):
    """ Simple wrapper to open a TCP connection to the
        daemon and connect to the host running the daemon.
        Returns a python socket object.
    """
    sk = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
    sk.connect((host,port))
    return sk

def sendxml(daemonsock, native):
    """ The first argument has to be a connected socket object,
        the second argument is a qlist object (list, dict,
        int, bool, string), that will be converted in a suitable
        XML representation and is sent to the daemon.
    """
    os.write(daemonsock.fileno(), py2qlist(native))
    return 0

def recvxml(daemonsock, debugdump=False):
    """ Read a XML document from a connected socket object,
        and convert it back to a native python object.
        If the conversion fails for some reason,
        an empty string will be returned.
    """
    done = 0
    buffer = ''
    retval = ''
    while not done:
        # Using a larger buffer speeds up XML processing!
        smallbuf = os.read(daemonsock.fileno(), 14 * 1024 * 1024)
        if len(smallbuf) == 0: # remote side closed
            done = 1
        buffer += smallbuf
        try:
            retval = qlist2py(buffer)
            done = 1
        except:
            retval = ''
    if debugdump:
        print "========================debugdump==========================="
        print buffer
        print "========================debugdump==========================="
    return retval

import threading
import time
            
class PrivateTimer (threading.Thread):
    stopme = False
    timeout = 100
    thesocket = ''

    def __init__(self, asocket, atimeout=None):
        super(PrivateTimer, self).__init__()
        self.thesocket = asocket
        self.timeout = atimeout
        self.stopme = False

        if self.timeout is None:
            self.timeout = sys.maxint

    def run(self):
        while True:
            if self.stopme:
                break
            if (self.timeout <= 0):
                #self.thesocket.close() ## shutdown is better, it shuts both sides.
                self.thesocket.shutdown(socket.SHUT_RDWR)
                break;
            time.sleep(1)
            self.timeout = self.timeout - 1

def timed_recvxml(daemonsock, timeout=None, debugdump=False):
    pt = PrivateTimer(daemonsock, timeout)
    pt.start()
    retval = recvxml(daemonsock, debugdump)
    pt.stopme = True
    pt.join()
    return retval

